#include <iostream>
#include <fstream>
#include <string>
#include <random>
#include <cstdlib>
#include <stdexcept>
#include <getopt.h>
#include "version.hpp"
using namespace std;

static double next_random() {
    static random_device dev;
    static default_random_engine engine {dev()};
    static uniform_real_distribution<double> dist {0, 1};
    return dist(engine);
}

static string random_line_from_file(istream& file) {
    // This function selects a line from a file by reading
    // the file line by line, so the file is never in memory
    // as a whole. Still, the chance a line is being picked
    // is equal for every line.
    string line;
    string result;
    for (int nr = 1; getline(file, line); ++nr) {
        if (next_random() * nr < 1.0) {
            result = line;
        }
    }
    return result;
}

static void print_help() {
    cout << "usage: ranlin [-h|--version]\n";
    cout << "  -h, --help    show this help text and exit\n";
    cout << "  --version     show version number and exit\n";
}

int main(int argc, char* argv[]) {
    try {
        int opt_char, opt_val;
        struct option long_options[] = {
            {"help", no_argument, nullptr, 'h'},
            {"version", no_argument, &opt_val, 1},
            {nullptr, 0, nullptr, 0}
        };
        while ((opt_char = getopt_long(argc, argv, "h", long_options, nullptr)) != -1) {
            string arg {optarg ? optarg : ""};
            switch (opt_char) {
                case 0: {
                    // handle long-only options here
                    switch (opt_val) {
                        case 1:
                            cout << ranlin_version() << endl;
                            return EXIT_SUCCESS;
                    }
                    break;
                }
                case 'h':
                    print_help();
                    return EXIT_SUCCESS;
                case '?':
                    throw runtime_error("unrecognized option");
            }
        }
        if (optind == argc) {
            // here when no file args
            cout << random_line_from_file(cin) << '\n';
        }
        for (int i = optind; i < argc; ++i) {
            try {
                // process file argv[i]
                ifstream infile {argv[1]};
                cout << random_line_from_file(infile) << '\n';
            } catch (const runtime_error& ex) {
                cerr << "ranlin: " << ex.what() << '\n';
            }
        }
    } catch (const exception& ex) {
        cerr << "ranlin: " << ex.what() << '\n';
        return EXIT_FAILURE;
    }
    return EXIT_SUCCESS;
}
